/*	@match.c - find matching pattern in buffer
 */
#include <stdio.h>
#include "gen.h"
#include "pat.h"

/*	return length of subpat in offset
 */
static Bool mtchccl(register char c, register char *pat, unsigned *offset)
	{
	register char rlc, ruc, *stpat = pat;
	Bool matched;

	for (matched = NO; *pat != P_CCLEND; )
		switch (*pat)
			{
		case P_CCHAR:
			if (c == *++pat)
				matched = YES;
			++pat;
			break;
		case P_RANGE:
			for (rlc = *++pat, ruc = *++pat; rlc <= ruc; rlc += 1)
				if (c == rlc)
					matched = YES;
			++pat;
			break;
			}
	*offset = pat - stpat + 1;
	return (matched);
	}

/* try to match one sub pattern
 */
static Bool omatch(char *buf, register unsigned n, register unsigned *pindx, char **pat, MSUB *pmsub)
	{
	register char *pp, *ppat;
	short bump;
	unsigned offset;

	pp = &buf[*pindx];
	ppat = *pat;
	bump = -1;
	switch (*ppat)
		{
	case P_SBOL:
		if (*pindx == 0)
			{
			bump = 0;
			ppat += 1;
			}
		break;
	case P_ANY:
		if (*pindx < n && *pp != '\n')
			bump = 1;
		ppat += 1;
		break;
	case P_SEOL:
		if (*pp == '\n' || *pindx == n)
			{
			bump = 0;
			ppat += 1;
			}
		break;
	case P_CCL:
		if (*pindx < n && mtchccl(*pp, ++ppat, &offset))
			bump = 1;
		ppat += offset;
		break;
	case P_NCCL:
		if (*pindx < n && *pp != '\n' && !mtchccl(*pp, ++ppat, &offset))
			bump = 1;
		ppat += offset;
		break;
	case P_LPAR:
		if (pmsub)
			{
			pmsub[ppat[1]].mtext = pp;
			pmsub[ppat[1]].mlen = 0;
			}
		ppat += 2;
		bump = 0;
		break;
	case P_RPAR:
		if (pmsub)
			pmsub[ppat[1]].mlen = pp - pmsub[ppat[1]].mtext;
		ppat += 2;
		bump = 0;
		break;
	case P_CCHAR:
		*pat += 2;
		if (*pindx < n && *pp == ppat[1])
			{
			++*pindx;
			return (YES);
			}
		return (NO);
		}
	*pat = ppat;
	if (0 <= bump)
		{
		*pindx += bump;
		return (YES);
		}
	else
		return (NO);
	}

/*	look for match anchored at index
 */
unsigned amatch(char *buf, register unsigned n, unsigned indx, char *pat, MSUB *pmsub)
	{
	register unsigned k;
	unsigned i;
	char *rpat;

	while (*pat != P_PEND)
		if (*pat == P_CLOSE)
			{
			for (rpat = pat + 1, i = indx; i < n; rpat = pat + 1)
				if (!omatch(buf, n, &i, &rpat, NULL))
					break;
			for (; ; --i)
				if ((k = amatch(buf, n, i, rpat, pmsub)) != -1 || indx == i)
					return (k);
			}
		else if (!omatch(buf, n, &indx, &pat, pmsub))
			return (-1);
	return (indx);
	}

/*	look for unanchored match
 */
Bool match(char *buf, register unsigned n, register char *pat)
	{
	register unsigned indx;

	if (*pat == P_SBOL)
		return (amatch(buf, n, 0, pat, NULL) != -1);
	for (indx = 0; indx < n; ++indx)
		if (buf[indx] != pat[3] && pat[2] == P_CCHAR)
			;
		else if (amatch(buf, n, indx, pat, NULL) != -1)
			return (YES);
	return (NO);
	}
